/*
 * replay.c:	Forward AC3 stream to the S/P-DIF interface of
 *		sound cards supported by ALSA. (NO decoding!)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include "types.h"
#include "replay.h"
#include "channel.h"
#include "ac3.h"
#include "dts.h"
#include "lpcm.h"
#include "pts.h"

// --- cReplayOutAC3 : Forward AC3 stream to S/P-DIF of a sound card with ALSA -----------

const uint_32 cReplayOutAC3::PS1magic = char2int(0x00,0x00,0x01,0xBD);
const uint_16 cReplayOutAC3::AC3magic = char2short(0x0b,0x77);
const uint_32 cReplayOutAC3::DTSmagic = char2int(0x7f,0xfe,0x80,0x01);
uint_8  * cReplayOutAC3::pesdata;
cBounce * cReplayOutAC3::bounce;

cReplayOutAC3::cReplayOutAC3(spdif &dev, ctrl_t &up, cBounce * bPtr, const char *script)
: SPDIFmute(script), spdifDev(&dev), setup(up)
{
    flags = 0;
    stream = NULL;
    bounce = bPtr;
    pesdata = setup.buf + TRANSFER_START;
}

cReplayOutAC3::~cReplayOutAC3(void)
{
    Clear();
}

void cReplayOutAC3::Activate(bool onoff)
{
    if (onoff) {
	if (*spdifDev)
	    goto out;		// An other has opened the S/P-DIF
	if (test_setup(RESET)) {
	    clear_flag(FAILED);
	    clear_setup(RESET);
	}
	if (test_flag(FAILED))
	    goto out;
	set_flag(ACTIVE);
	bounce->bank(0);
	Start();
	pthread_yield();
    } else {
	if (!test_flag(RUNNING))
	    goto out;		// An other has opened the S/P-DIF
	clear_flag(ACTIVE);
	while (test_flag(RUNNING)) {
	    bounce->flush();
	    bounce->signal();
	    usleep(1000);
	}
	Cancel(1);
    }
out:
    return;
}

void cReplayOutAC3::Action(void)
{
    set_flag(RUNNING);
    iec60958* tmp;
    if (!stream || !spdifDev->Open(stream, setup)) {
	esyslog("REPLAY: can't open spdif interface");
	set_flag(FAILED);
	goto out;
    }

    dsyslog("REPLAY: Forwarding bitstream thread started (pid=%d)", getpid());
    realtime("INSTREAM: ");
    bounce->takeio();
    tmp = stream;
    while (test_flag(ACTIVE)) {
	ssize_t len;

	if (test_setup(MUTE))
	    spdifDev->Clear();

	if (!spdifDev->Synchronize(bounce))	// Wait on spdifDev->Signal()
	    break;

	if (test_setup(MUTE)) {
	    bounce->flush();
	    bounce->bank(0);
	    clear_flag(BOUNDARY);
	    continue;
	}

	if ((len = bounce->fetch(pesdata, spdifDev->Available(TRANSFER_MEM)))) {
	    /* int count = */ spdifDev->Forward(pesdata, len, bounce);
	}

    }
    spdifDev->Close();
    dsyslog("REPLAY: Forwarding bitstream thread ended (pid=%d)", getpid());
    tmp->Reset();
out:
    clear_flag(RUNNING);
    stream = NULL;
    bounce->leaveio();
}

inline bool cReplayOutAC3::OffsetToPayOfPS1(const uchar *const b, int &off, const int cnt)
{
    bool ret = true;				// DVB streams do not have sub stream headers
    const uchar *const p = &b[off];

    // Should we check the stream number (p[0] & 0x07)?
    // Normally this plugin gets only one audio stream.

    switch ((p[0] & 0xf8)) {
    case 0x80:					// AC3 from e.g. DVD
	if (stream != &ac3 || stream->track != (p[0] & 7))
	    ret = false;
	off += 4;
	break;
    case 0x88:					// DTS
	if (stream != &dts || stream->track != (p[0] & 7))
	    ret = false;
	off += 4;
	break;
    case 0xA0:					// lPCM
	if (stream != &pcm || stream->track != (p[0] & 7))
	    ret = false;
	off += 7;
	break;
    default:					// Should not happen, e.g. stream
	ret = false;				// is broken. Should we reset???
	break;
    }

    if (cnt < off+1)
	ret = false;

    return ret;
}

inline bool cReplayOutAC3::ScanPayOfPS1(const uchar *const b, int &off, const int cnt)
{
    const uchar *const p = &b[off];
    const uchar *m;

    if (stream)
	goto out;

#define available(x)	if (cnt < off+(x)) goto out

    // Should we check the stream number (p[0] & 0x07)?
    // Normally this plugin gets only one audio stream.

    switch ((p[0] & 0xf8)) {
    case 0x80:					// AC3 from e.g. DVD
	available(4);
	off += (((uint_16)p[2]<<8)|p[3]) + 3;
	available(2);

	m = &b[off];
	if (char2short(m[0],m[1]) == AC3magic) {
	    stream = &ac3;
	    stream->Reset();
	    stream->isDVD = true;
	    stream->track = (p[0] & 7);
	    clear_setup(AUDIO);
	}

	break;
    case 0x88:					// DTS
	available(4);
	off += (((uint_16)p[2]<<8)|p[3]) + 3;
	available(4);

	m = &b[off];
	if (char2int(m[0],m[1],m[2],m[3])  == DTSmagic) {
	    stream = &dts;
	    stream->Reset();
	    stream->isDVD = true;
	    stream->track = (p[0] & 7);
	    clear_setup(AUDIO);
	}

	break;
    case 0xA0:					// lPCM
	available(8);
	if ((((uint_16)p[2]<<8)|p[3]) > 3)	// Offset to first channel 0 (left) sample
	    off += (((uint_16)p[2]<<8)|p[3]) + 3;
	else
	    off += 7;
	available(1);

	stream = &pcm;
	stream->Reset();
	stream->isDVD = true;

	switch (p[5] & 0x30) {
	case 0x00: pcm.sample_rate = 48000;
	    break;
	case 0x20: pcm.sample_rate = 44100;
	    break;
	case 0x30: pcm.sample_rate = 32000;
	    break;
	case 0x10:				// 96kHz currently not supported
	default:   stream = NULL;
	    break;
	}
	switch (p[5] & 0xC0) {
	case 0x00:				// 16 bit samples
	    break;
	case 0x40:				// 20 bit currently not supported
	case 0x80:				// 24 bit currently not supported
	default:   stream = NULL;
	    break;
	}
	switch (p[5] & 0x07) {
	case 0x00:				// 1 channel
	case 0x01:				// 2 channels
	    break;
	default:   stream = NULL;		// More channels not supported
	    break;
	}

	if (!stream)
	    goto out;
	set_setup(AUDIO);

	break;
    default:					// e.g. AC3 emmbedded within DVB PS1
	available(2);
	if (char2short(p[0],p[1]) == AC3magic) {
	    stream = &ac3;
	    stream->Reset();
	    stream->isDVD = false;
	    stream->track = (p[0] & 7);
	    clear_setup(AUDIO);
	}
	break;
    }

#undef available

out:
    return (stream != NULL);
}

void cReplayOutAC3::Play(const uchar *b, int cnt)
{
    if (test_setup(MUTE))
	goto out;

    if (!test_setup(ACTIVE)) {
	Clear();
	goto out;
    }

    if (cnt < 10)
	goto out;

    if ((b[6] & 0xC0) != 0x80)
	goto out;			// Not PS1 Mpeg2 PES

    if (char2int(b[0],b[1],b[2],b[3]) == PS1magic) {
	int len = (((uint_16)b[4]<<8)|b[5]) + 6;
	int off = b[8] + 9;

	if (cnt < off+1)
	    goto out;

	if (!test_flag(BOUNDARY)) {

	    // If the stream provides PTS/DTS information and/or
	    // the data in the pay load is marked as aligned we
	    // assume that that we'll find immediately the sync words
	    // of the pay load its self or the corresponding sub
	    // stream identification with the offset to the pay load.
	    
	    if ((b[7] & 0xC0) || (b[6] & 0x04)) {

		if (!ScanPayOfPS1(&b[0], off, cnt))
		    goto out;

		set_flag(BOUNDARY);	// At the beginning of a stream

		if (!test_flag(RUNNING))
		    Activate(true);
#if 0
		if (b[7] & 0xC0) {
		    // Do we need a sanity check? If so we have to
		    // use a global variable count (locking!)
		    // for increasing it here and decreasing it
		    // at Action()
		    XXXX
	        }
#endif
	    } else
		goto out;		// Start only at a boundary of a stream

	} else if (stream && stream->isDVD) {

	    if (!OffsetToPayOfPS1(b, off, cnt)) {
		Clear();		// Reset
		goto out;
	    }

	}
#ifndef SPDIF_TEST
	bounce->store((uint_8 *)&b[off], min(len - off, cnt));
#else
fprintf(stderr, "store(0x%.2x 0x%.2x 0x%.2x 0x%.2x, %d)\n", b[off],b[off+1],b[off+2],b[off+3], min(len - off, cnt));
frame_t pcm;
const uint_8 *end = (uint_8 *)&b[off] + min(len - off, cnt);
const uint_8 *data = (uint_8 *)&b[off];
while ((pcm = stream->Frame(data, end)).burst) {
	;
}
#endif
    }
out:
    return;
}

void cReplayOutAC3::Mute(bool onoff)
{
    if ((test_setup(MUTE) ? true : false) == onoff)
	goto out;

    if (onoff) {
	set_setup(MUTE);
	spdifDev->Pause(onoff);
    } else
	clear_setup(MUTE);
    bounce->signal();

    if (SPDIFmute) {	// Normal TV mode
	char *cmd = NULL;
	asprintf(&cmd, "%s %s 2> /dev/null", SPDIFmute, test_setup(MUTE) ? "mute" : "unmute");
	if (cmd) {
	    system(cmd);
	    free(cmd);
	    cmd = NULL;
	}
    }
out:
    return;
}

void cReplayOutAC3::Clear(void)
{
    (void)Activate(false);
    stream = NULL;
    clear_flag(BOUNDARY);
#ifndef SPDIF_TEST
    Mute(cDevice::PrimaryDevice()->IsMute());
#endif
}

#ifdef SPDIF_TEST
# include <sys/time.h>
# include <sys/types.h>
# include <sys/ioctl.h>

static ctrl_t setup = {
    NULL,       // buf
    {
        0,      // opt.card
        2,      // opt.device
        16,     // opt.delay
        false,  // opt.type
        true,   // opt.variable
        false   // opt.mmap
    },
    false,      // audio
    false,      // reset
    false,      // mute
    true        // active
};

static spdif spdifDev;
static cBounce * bounce;
static char *SPDIFmute;

int
main ()
{
    uchar buffer[10*2048];
    cReplayOutAC3  *ReplayOutAC3 = NULL;

    if (!(setup.buf = (uint_8 *)malloc(sizeof(uint_8)*OVERALL_MEM)))
        exit(-1);

    bounce = new cBounce(setup.buf, BOUNCE_MEM);
    ac3.SetBuffer(setup.buf + SPDIF_START);
    dts.SetBuffer(setup.buf + SPDIF_START);
    pcm.SetBuffer(setup.buf + SPDIF_START);

    ReplayOutAC3 = new cReplayOutAC3(spdifDev, &setup, bounce, SPDIFmute);

    uchar * tail = &buffer[0];
    uchar * head = &buffer[0];
    uchar * end  = &buffer[0] + sizeof(buffer);
    while (1) {

	ssize_t chunk = 0;
	size_t rest = end - tail;
	do {

	    if ((chunk = read(0, tail, rest)) < 0) {
		if (errno == EINTR) {
		    usleep(10);
		    continue;
		}
		fprintf(stderr, "read error: %s\n", strerror(errno));
		exit(1);
	    }
	    if (chunk) {
		rest -= chunk;
		tail += chunk;
	    }

	} while (chunk > 0 && rest > 0);

	int warnme = 1;
	size_t got = tail - head;
	while (got > 6) {

	    if (head[0] == 0 && head[1] == 0 && head[2] == 1 && (head[3] == 0xBD || head[3] == 0xBE)) {
		size_t len = (((uint_16)head[4]<<8)|head[5]) + 6;

		warnme = 1;

		if (got < len)
		    break;

		ReplayOutAC3->Play(head, len);
		head += len;
		got  -= len;

	    } else {

		if (warnme)
		    fprintf(stderr, "\n ###STREAM BROKEN %d %d\n\n", chunk, got);
		warnme = 0;

		head += 1;
		got  -= 1;
	    }
	}

	if (head > &buffer[0]) {
	    size_t avail = tail - head;
	    head = (uchar*)memmove(&buffer[0], head, avail);
	    tail = head + avail;
	}

	if (!chunk)
	    break;
    }
}
#endif
