/*
 * xlist.c:	List the content of MPEG files
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <netinet/in.h>

#ifndef AARONS_TYPES
#define AARONS_TYPES
typedef unsigned long long uint_64;
typedef unsigned int   uint_32;
typedef unsigned short uint_16;
typedef unsigned char  uint_8;

typedef signed long long sint_64;
typedef signed int     sint_32;
typedef signed short   sint_16;
typedef signed char    sint_8;
#endif

#define dsyslog(format, args...)	fprintf(stderr, "xlist: " format "\n\r", ## args)
#define O_PIPE				(O_NONBLOCK|O_ASYNC)

// Exeption class
class cExeption {
public:
    const int reason;
    cExeption(const int Reason) : reason(Reason) {};
};

// Handle class
class cHandle {
private:
    const int fd;
    uint_64 offset;
    uint_8* const data;
    uint_8* head;
    uint_8* tail;
    const uint_8* const end;
    ssize_t left;
    cExeption ex;
    virtual bool refill(off_t off = 0) throw(cExeption);
    virtual inline void bounce(void);
public:
    cHandle(uint_8* const buf, const size_t len, const int FD = -1)
    : fd(FD), data(buf), head(buf), tail(buf), end(buf+len), ex(1)
    {
	offset = 0;
	if (fd < 0)
	    tail += len;
	refill(1);
    };
    virtual ~cHandle(void) {};
    virtual const uint_64 Offset() const { return offset; };
    operator uint_8  () { return refill(1) ? *head : 0; };
    operator uint_16 () { return refill(2) ? ntohs(*((uint_16*)(&head[0]))) : 0; };
    operator uint_32 () { return refill(4) ? ntohl(*((uint_32*)(&head[0]))) : 0; };
    operator bool    () { return refill(1); };
    uint_8* operator () (const off_t o) { return refill(o) ? head : NULL; };
    cHandle& operator ++()
    {
	if (refill(1)) { head++; offset++; }
	return *this;
    };
    cHandle operator ++(int)
    {
	bounce();		// Before copying pointers move head to data start
	const cHandle ret(*this);
	if (refill(1)) { head++; offset++; } 
	return ret;
    };
    cHandle& operator +=(const off_t o)
    {
	if (refill(o)) { head +=o; offset+=o; }
	return *this;
    };
    virtual size_t len() { return refill(1) ? (size_t)left : 0; };
#   define foreach(handle)	try { while (handle)
#   define test(handle)		try { if(handle)
#   define end(handle)		} catch (cExeption& ex) { if (ex.reason != 1) throw; }
};

inline void cHandle::bounce(void)
{
    if (head <= data)
	return;
    left = tail - head;
    head = (uint_8*)memmove(data, head, left);
    tail = head + left;
}

// This function has to be called _before_ head is read
bool cHandle::refill(off_t off)	throw(cExeption)
{
    if (fd < 0) 	// For handle static buffer only
	goto out;

    if ((tail - head) > (end - data)/2)
	goto out;

    bounce();

    // XXX: Should we use a ringbuffer with overlay area?
    do {
	ssize_t len = read(fd, tail, end - tail);
	if (len < 0) {
	    if (errno == EINTR || errno == EAGAIN)
		continue;
	    dsyslog("cHandle::refill(): %s", strerror(errno));
	    throw ex;
	}
	if (len  > 0)
	    tail += len;
        if (len == 0)
	    break;

    } while (end - tail);
out:
    if ((left = tail - head) >= off)
	return true;
    else
	if (left > 0) 
	    head += left;
    throw ex;
}

// Signal handling
static struct sigaction saved_act[SIGUNUSED];
static void sighandler(int sig)
{
    sigset_t set;

    if (sig != SIGPIPE && sig != SIGIO)
	dsyslog("Catch %s", strsignal(sig));

    if (sig == SIGIO)
	return;

    usleep(100);

    if (sigaction (sig, &saved_act[sig], NULL))
	_exit(1);
    if (sigemptyset(&set) || sigaddset(&set, sig) ||
	sigprocmask(SIG_UNBLOCK, &set, NULL))
	_exit(1);
    kill(-getpid(), sig);	// Raise signal
}

static void install_sighandler(int sig)
{
    struct sigaction act;
    if (sigaction (sig, NULL, &saved_act[sig]))		// Old action handler
	return;
    if (saved_act[sig].sa_handler == sighandler)	// Compare with our
	return;

    act.sa_handler = sighandler;
    if (sig == SIGIO) {
	act.sa_flags = SA_RESTART;
    } else {
	act.sa_flags = SA_ONESHOT;
    }
    if (sigemptyset (&act.sa_mask) < 0)
	_exit(1);
    if (sigaction (sig, &act, NULL))
	_exit(1);
}

// Known options
static FILE* out = stdout;
#define D(format, args...)		fprintf(out, format "\n", ## args)
#define N(format, args...)		fprintf(out, format,      ## args)
#define nl				putc('\n', out);
static struct option long_option[] =
{
    {"help",      0, NULL,  'h'},
    {"block",     0, NULL,  'b'},
    {"output",    1, NULL,  'o'},
    {"noslices",  1, NULL,  's'},
    {"nobdpay",   1, NULL,  'd'},
    { NULL,       0, NULL,   0 },
};

static void help(void)
{
        printf("Usage: xlist <options> [file]\n");
        printf("\nAvailable options:\n");
        printf("  -h, --help         this help\n");
        printf("  -b, --block        use blocking read mode on stdin\n");
        printf("  -o, --output=file  use this file for output\n");
        printf("  -s, --noslices     do not show video slice nor sequences\n");
        printf("  -d, --nobdpay      do not show payload of private stream 1\n");
}

// Forward declaration
static bool skip_slices = false;
static bool skip_bdpayl = false;
static void scan(cHandle handle);

int main(int argc, char *argv[])
{
    bool block = false;
    cHandle *handle;
    char *output;
    int c;

    while ((c = getopt_long(argc, argv, "bo:hsd", long_option, NULL)) > 0) {
	switch (c) {
	case 'b':
	    block = true;
	    break;
	case 'o':
	    if (!optarg || *optarg == '-') {
		help();
		exit(1);
	    }
	    if (!(output = strdup(optarg))) {
		dsyslog("strdup() failed: %s", strerror(errno));
		exit(1);
	    }
	    if (!(out = fopen(output, "a+"))) {
		dsyslog("fopen(%s) failed: %s", output, strerror(errno));
		exit(1);
	    }
	    break;
	case 'h':
	    help();
	    exit(0);
	    break;
	case 's':
	    skip_slices = true;
	    break;
	case 'd':
	    skip_bdpayl = true;
	    break;
	default:
	    help();
	    exit(1);
	    break;
	}
    }
    argv += optind;
    argc -= optind;

    if (*argv && **argv == '-') {
	argv++;
	argc--;
    }

    install_sighandler(SIGINT);
    install_sighandler(SIGQUIT);
    install_sighandler(SIGHUP);
    install_sighandler(SIGTERM);
    install_sighandler(SIGPIPE);
    install_sighandler(SIGIO);

    bool loop = true;
    static uint_8 buf[32*4096];

    while (true) {
	int fd = STDIN_FILENO;

	if (argc <= 0) {
	    if (!loop)
		break;
	    if (!block && fd == STDIN_FILENO) {
		int flags = fcntl(fd, F_GETFL);
		if (flags >= 0) {
		    if (fcntl(fd, F_SETFL, (flags|O_PIPE)) < 0 ||
			fcntl(fd, F_SETOWN, getpid()) < 0)
			dsyslog("Couldn't set NONBLOCK|ASYNC mode on stdin: %s", strerror(errno));
		}
	    }
	} else {
	    if (!*argv)
		break;
	    if ((fd = open(*argv, O_NOCTTY|O_RDONLY)) < 0) {
		dsyslog("Couldn't open file %s: %s", *argv, strerror(errno));
		exit(1);
	    }
	    argv++;
	    argc--;
	}
	loop = false;

	handle = new cHandle(&buf[0], sizeof(buf), fd);

	scan(*handle);

	if (handle) {
	    delete handle;
	    handle = NULL;
	}

	if (fd != STDIN_FILENO)
	    close (fd);
    }

    return 0;
}

// Used in scan for video data
static off_t scan_e0(uint_8* const buf, const off_t rest, const uint_64 deep);

// Used in scan for private stream data
#define BD_IS_ALIGNED	0x01
#define BD_HAS_PTSDTS	0x02
static off_t scan_bd(uint_8* const buf, const off_t rest, const uint_8 flags, const uint_64 deep);


static void scan(cHandle handle)
{
    char I_Frame = 0; 
    bool Header  = false;

    foreach(handle(4)) {
	uint_32 ul = handle;
	if ((ul & 0xffffff00) == 0x0000100) {
	    uint_64 offset = handle.Offset();
	    handle += 4;

	    uint_8  ub = (uint_8)(ul & 0x000000ff);
	    uint_16 m  = handle;
	    uint_8  l  = (uint_8)((m>>8)&0x00ff);

	    N("%8llu %02X", offset, ub);
	    switch (ub) {
	    case 0xb9:		// Program end (terminates a program stream)
		D(" program_end");
		break;
	    case 0xba:		// Pack header
		if (((l>>2) & 0x01) != 1 && (l>>6) != 1)
		    break;
		Header = true;
		I_Frame = 0;
		N(" pack_header");
		for (int i = 0; i < 10; i++)
		    N(" %02x", (uint_8)handle++);
		nl;
		break;
	    case 0xbb:		// System Header
		handle += 2;
		N(" system_header");
		N(", len=%d", m+6);
		N(", ext:");
		for (int i = 0; i < m && i < 20; i++)
		    N(" %02x", (uint_8)handle++);
		nl;
		break;
	    case 0xbc:		// Program Stream Map
		D(" program_stream_map");
		break;
	    case 0xbd:		// Private stream 1
		handle += 2;
		N(" private_stream_1");
		N(", len=%d", m+6);
		{
		    uint_8 *p;
		    uint_8  h;
		    uint_8  b = 0;
		    uint_16 l = m;

		    if (m < 2)
			break;
		    if (m > handle.len())
			m = handle.len();
		    if (!(p = handle(m)))
			break;
		    if (m < (h = p[2] + 3))
			break;
		    l -= h;

		    if ((p[0] & 0xC0) != 0x80) {
			N(" (broken)");
			break;
		    }

		    if (p[0]&0x04)
			b |= BD_IS_ALIGNED;
		    if (p[1] & 0xC0)
			b |= BD_IS_ALIGNED;

		    off_t off;
		    N(", ext:");
		    for (off = 0; off < h; off++)
			N(" %02x", p[off]);

		    if ((p[1] & 0x80) && (p[2] >= 5)) {
			static uint_64 lpts = 0;
			uint_64 pts = 0;
			pts  = (((uint_64)p[3]) & 0x0e) << 29;
			pts |= ( (uint_64)p[4])         << 22;
			pts |= (((uint_64)p[5]) & 0xfe) << 14;
			pts |= ( (uint_64)p[6])         <<  7;
			pts |= (((uint_64)p[7]) & 0xfe) >>  1;
			if (lpts) {
			    sint_64 diff = pts - lpts;
			    N(", pts: %llu          %lld", pts, diff);
			    if (diff > 0)
				lpts = pts;
			} else {
			    N(", pts: %llu", pts);
			    lpts = pts;
			}
		    }

		    if (skip_bdpayl)
			handle += m;
		    else
			handle += off + scan_bd(p+off, m-off, b, handle.Offset()+off);
		}
		nl;
		break;
	    case 0xbe:		// Padding stream
		handle += 2;
		l = handle;
		if (l != 0xff)
		    break;
		handle += m;
		D(" padding_stream, len=%d", m+6);
		break;
	    case 0xbf:		// Private stream 2
		handle += (2+m);
		D(" private_stream_2, len=%d", m+6);
		break;
	    case 0xc0 ... 0xdf:	// MPEG-1 or MPEG-2 audio stream
		handle += 2;
		N(" audio_stream(%d)", (ub & 0x1f));
		N(", len=%d", m+6);
		{
		    uint_8  *p;
		    uint_8  h;

		    if (m < 2)
			break;
		    if (m > handle.len())
			m = handle.len();
		    if (!(p = handle(m)))
			break;
		    if (m < (h = p[2] + 3))
			break;

		    if ((p[0] & 0xC0) != 0x80) {
			N(" (broken)");
			break;
		    }

		    off_t off;
		    N(", ext:");
		    for (off = 0; off < h; off++)
			N(" %02x", p[off]);

		    if ((p[1] & 0x80) && (p[2] >= 5)) {
			static uint_64 lpts = 0;
			uint_64 pts = 0;
			pts  = (((uint_64)p[3]) & 0x0e) << 29;
			pts |= ( (uint_64)p[4])         << 22;
			pts |= (((uint_64)p[5]) & 0xfe) << 14;
			pts |= ( (uint_64)p[6])         <<  7;
			pts |= (((uint_64)p[7]) & 0xfe) >>  1;
			if (lpts) {
			    sint_64 diff = pts - lpts;
			    N(", pts: %llu          %lld", pts, diff);
			    if (diff > 0)
				lpts = pts;
			} else {
			    N(", pts: %llu", pts);
			    lpts = pts;
			}
		    }

		    handle += m;
		}
		nl;
		break;
	    case 0xe0 ... 0xef:	// MPEG-1 or MPEG-2 video stream
		handle += 2;
		N(" video_stream(%d)", (ub & 0x0f));
		N(", len=%d", m+6);
		{
		    uint_8 *p;
		    uint_8 h;

		    if (m < 2)
			break;
		    if (m > handle.len())
			m = handle.len();
		    if (!(p = handle(m)))
			break;
		    if (m < (h = p[2] + 3))
			break;

		    if ((p[0] & 0xC0) != 0x80) {
			N(" (broken)");
			break;
		    }

		    off_t off;
		    N(", ext:");
		    for (off = 0; off < h; off++)
			N(" %02x", p[off]);

		    if ((p[1] & 0x80) && (p[2] >= 5)) {
			static uint_64 lpts = 0;
			uint_64 pts = 0;
			pts  = (((uint_64)p[3]) & 0x0e) << 29;
			pts |= ( (uint_64)p[4])         << 22;
			pts |= (((uint_64)p[5]) & 0xfe) << 14;
			pts |= ( (uint_64)p[6])         <<  7;
			pts |= (((uint_64)p[7]) & 0xfe) >>  1;
			if (lpts) {
			    sint_64 diff = pts - lpts;
			    N(", pts: %llu          %lld", pts, diff);
			    if (diff > 0)
				lpts = pts;
			} else {
			    N(", pts: %llu", pts);
			    lpts = pts;
			}
		    }

		    if (skip_slices)
			handle += m;
		    else
			handle += off + scan_e0(p+off, m-off, handle.Offset()+off);
		}
		nl;
		break;
	    case 0xf0:		// ECM Stream
		D(" ECM Stream");
		break;
	    case 0xf1:		// EMM Stream
		D(" EMM Stream");
		break;
	    case 0xf2:		// ITU-T Rec. H.222.0
				// ISO/IEC 13818-1 Annex A or ISO/IEC 13818-6 DSMCC stream
		D(" ITU-T Rec. H.222.0 | ISO/IEC 13818-1 | ISO/IEC 13818-6 DSMCC");
		break;
	    case 0xf3:		// ISO/IEC_13522_stream
		D(" ISO/IEC_13522_stream");
		break;
	    case 0xf4:		// ITU-T Rec. H.222.1 type A
		D(" ITU-T Rec. H.222.1 type A");
		break;
	    case 0xf5:		// ITU-T Rec. H.222.1 type B
		D(" ITU-T Rec. H.222.1 type B");
		break;
	    case 0xf6:		// ITU-T Rec. H.222.1 type C
		D(" ITU-T Rec. H.222.1 type C");
		break;
	    case 0xf7:		// ITU-T Rec. H.222.1 type D
		D(" ITU-T Rec. H.222.1 type D");
		break;
	    case 0xf8:		// ITU-T Rec. H.222.1 type E
		D(" ITU-T Rec. H.222.1 type E");
		break;
	    case 0xf9:		// ancillary_stream
		D(" ancillary_stream");
		break;
	    case 0xff:		// Program Stream Directory
		D(" program_stream_directory");
		break;
	    default:
		D(" unkown or reserved %08x", ul);
		break;
	    }
	} else {
	    handle++;
	}
    } end (handle);
};

static off_t scan_e0(uint_8* const buf, const off_t rest, const uint_64 deep)
{
    cHandle handle(buf, rest);
    char I_Frame = 0; 

    foreach(handle(4)) {
	uint_32 ul = handle;
	if ((ul & 0xffffff00) == 0x0000100) {
	    uint_64 offset = handle.Offset() + deep;
	    handle += 4;

	    uint_8  ub = (uint_8)(ul & 0x000000ff);
	    uint_32 n  = handle;
	    uint_16 m  = (uint_16)((n>>16)&0x0000ffff);
	    uint_8  l  = (uint_8) ((m>> 8)&0x00ff);

	    nl;
	    N("%8llu %02X", offset, ub);
	    switch (ub) {
	    case 0x00:		// Picture
		handle += 4;
		switch ((m>>3)&0x07) {
		case 1:  I_Frame = 'I'; break;
		case 2:  I_Frame = 'P'; break;
		case 3:  I_Frame = 'B'; break;
		case 4:  I_Frame = 'D'; break;
		default: I_Frame =  0 ; break;
		}
		N(" picture_header - temporal_reference: %4d  picture_coding_type: %c", m>>6 , I_Frame);
		break;
	    case 0x01 ... 0xaf:
		N(" slice");
		for (int i = 0; i < 20; i++) {
		    uint_32 n = handle;
		    if ((n & 0xffffff00) == 0x0000100 && n != ul)
			break;
		    N(" %02x", (uint_8)handle++);
		}
		break;
	    case 0xb2:		// User data
		N(" user_data_start");
		break;
	    case 0xb3:		// Sequence header
		handle += 11;
		N(" sequence_header");
		break;
	    case 0xb4:		// Sequence error
		N(" sequence_error");
		break;
	    case 0xb5:		// Extension
		N(" extension_start");
		switch (l>>4) {
		case 1:
		    N(" (sequence_extension)");
		    handle +=  6;
		    break;
		case 2:
		    N(" (sequence_display_extension)");
		    handle++;
		    l = handle;	
		    if (l & 1)
			handle += 12;
		    else
			handle += 5;
		    break;
		case 8:
		    N(" (picture_coding_extension)");
		    handle += 4;
		    l = handle;
		    if (l & 0x40)
			handle += 3;
		    else
			handle++;
		    break;
		default:
		    break;
		}
		break;
	    case 0xb7:		// Sequence end
		N(" sequence_end");
		break;
	    case 0xb8:		// Group of Pictures
		N(" group_of_pictures_header - time_code:");
		N(" %02u:%02u:%02u.%02u", (n>>26)&0x1f, (n>>20)&0x3f, (n>>13)&0x3f, (n>>7)&0x3f);
		N(" closed_gop: %d broken_link: %d", (n>>6)&0x01, (n>>5)&0x01);
		break;
	    default:
		N(" unkown or reserved %08x", ul);
		break;
	    }
	} else {
	    handle++;
	}
    } end (handle);
    return rest;
}

static off_t scan_bd(uint_8* const buf, const off_t rest, const uint_8 flags, const uint_64 deep)
{
    cHandle handle(buf, rest);
    bool substream = false;

    test(handle(4)) {
	uint_32 ul = handle;
	uint_8  ub = (uint_8) (ul>>24);

	uint_8 c = (uint_8)((ul>> 16)&0x00ff);
	off_t  o = (off_t)(ul & 0x0000ffff) + 3;

	uint_16 m;
	uint_32 n;

	switch (ub) {
	case 0x20 ... 0x3f:	// Subpictures
	    handle++;
	    {
		static sint_32 len = 0;
		if (!flags && !len)
		    break;
		if (flags) {
		    substream = true;
		    len = (uint_16)handle;
		    nl;
		    N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		    N(" subpicture, len=%d spuh:", len);
		    for (int i = 0; i < 4; i++)
			N(" %02x", (uint_8)handle++);
		    len -= (rest - 1);
		} else if (len) {
		    substream = true;
		    nl;
		    N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		    N(" subpicture, rest=%d", len);
		    if ((len -= (rest - 1)) < 0)
			len = 0;
		}
	    }
	    break;
	case 0x80 ... 0x87:	// Dolby Digital Audio
	    if (o >= rest)
		break;
	    handle += o;
	    m = handle;
	    if (m == 0x0b77) {
		substream = true;
		nl;
		N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		N(" AC3(DVD), stream: %u, count: %u, start: %lu", (ub & 7), c, o - 3);
	    }
	    break;
	case 0x88 ... 0x8f:	// DTS Audio
	    if (o+2 >= rest)
		break;
	    handle += o;
	    n = handle;
	    if (n == 0x7ffe8001) {
		substream = true;
		nl;
		N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		N(" DTS(DVD), stream: %u, count: %u, start: %lu", (ub & 7), c, o - 3);
	    }
	    break;
	case 0xa0 ... 0xa7:	// Linear PCM Audio
	    if (o == 3)
		o = 7;
	    if (o >= rest)
		break;
	    handle += o;
	    {
		static sint_32 len = 0;
		if (!flags && !len)
		    break;
		if (flags) {
		    substream = true;
		    len = (rest - o);
		    nl;
		    N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		    N(" linear PCM, len=%d:", len);
		} else if (len) {
		    substream = true;
		    len = (rest - o);
		    nl;
		    N("%8llu %02X", deep+handle.Offset(), (ub & 0xf8));
		    N(" linear PCM, rest=%d", len);
		}
	    }
	    break;
	default:
	    break;
	}
    } end (handle);

    if (!substream) {
	cHandle handle(buf, rest);
	typedef struct _frmsize
	{
	    uint_16 bit_rate;
	    uint_16 frame_size[3];
	} frmsize_t;

	static const frmsize_t frmsizecod_tbl[64] = {
	    { 32  ,{64   ,69   ,96   } },
	    { 32  ,{64   ,70   ,96   } },
	    { 40  ,{80   ,87   ,120  } },
	    { 40  ,{80   ,88   ,120  } },
	    { 48  ,{96   ,104  ,144  } },
	    { 48  ,{96   ,105  ,144  } },
	    { 56  ,{112  ,121  ,168  } },
	    { 56  ,{112  ,122  ,168  } },
	    { 64  ,{128  ,139  ,192  } },
	    { 64  ,{128  ,140  ,192  } },
	    { 80  ,{160  ,174  ,240  } },
	    { 80  ,{160  ,175  ,240  } },
	    { 96  ,{192  ,208  ,288  } },
	    { 96  ,{192  ,209  ,288  } },
	    { 112 ,{224  ,243  ,336  } },
	    { 112 ,{224  ,244  ,336  } },
	    { 128 ,{256  ,278  ,384  } },
	    { 128 ,{256  ,279  ,384  } },
	    { 160 ,{320  ,348  ,480  } },
	    { 160 ,{320  ,349  ,480  } },
	    { 192 ,{384  ,417  ,576  } },
	    { 192 ,{384  ,418  ,576  } },
	    { 224 ,{448  ,487  ,672  } },
	    { 224 ,{448  ,488  ,672  } },
	    { 256 ,{512  ,557  ,768  } },
	    { 256 ,{512  ,558  ,768  } },
	    { 320 ,{640  ,696  ,960  } },
	    { 320 ,{640  ,697  ,960  } },
	    { 384 ,{768  ,835  ,1152 } },
	    { 384 ,{768  ,836  ,1152 } },
	    { 448 ,{896  ,975  ,1344 } },
	    { 448 ,{896  ,976  ,1344 } },
	    { 512 ,{1024 ,1114 ,1536 } },
	    { 512 ,{1024 ,1115 ,1536 } },
	    { 576 ,{1152 ,1253 ,1728 } },
	    { 576 ,{1152 ,1254 ,1728 } },
	    { 640 ,{1280 ,1393 ,1920 } },
	    { 640 ,{1280 ,1394 ,1920 } }
	};

	test(handle(4)) {
	    static sint_32 len = 0;
	    sint_32 l = rest;
	    uint_16 m;
	    if (flags || len) {
		if (flags)
		    len = 0;		// On boundary normally AC3 starts
		if (len) {
		    handle += len;	// If not on boundary a overroll may occured
		    l -= len;
		}
		while (l > 0) {
		    m = handle;
		    if (m == 0x0b77) {
			uint_16 fscod, frmsizecod;
			handle += 4;

			fscod	   = (((uint_8)handle)>>6)&3;
			frmsizecod =  ((uint_8)handle)&0x3f;
			len = 2*frmsizecod_tbl[frmsizecod].frame_size[fscod];

			nl;
			N("%8llu   ", deep+handle.Offset());
			N(" AC3(DVB) size=%d", len);
			if (l < len) {
			    len -= l;
			    break;
			}
			l -= len;
		    } else {
			len = 0;
			break;
		    }
		}
	    }
	} end (handle);
    }

    return rest;
}
